/*************************************************************************
 *
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 * 
 * Copyright 2008 by Nakata, Maho
 *
 * MPACK - multiple precision arithmetic library
 *
 * This file is part of MPACK.
 *
 * MPACK is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 3
 * only, as published by the Free Software Foundation.
 *
 * OpenOffice.org is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License version 3 for more details
 * (a copy is included in the LICENSE file that accompanied this code).
 *
 * You should have received a copy of the GNU Lesser General Public License
 * version 3 along with OpenOffice.org.  If not, see
 * <http://www.openoffice.org/license.html>
 * for a copy of the LGPLv3 License.
 *
 ************************************************************************/

using namespace std;

#define max(x,y) (x>y?x:y)

#define GMP_P_FORMAT "%21.16Fe"
#define P_FORMAT "%21.16e"
#define EPSILON  1e-13
#define EPSILON2 1e-10

#define TRUE     1
#define FALSE    0

#define VEC_A(n,inc) max(1,(1+(n-1)*abs(inc)))
#define MAT_A(lda,n) max(1,lda*n)

int VERBOSE=0;

gmp_randclass *uniformrandom;

//prototype
mpf_class mpf_randomnumber(void);
mpc_class mpc_randomnumber(void);
void setzerovec (int N, mpf_class *p);
void setrandvec (int N, mpf_class *p);
void setrandmat (int N, int M, mpf_class *p);
void printmat (int N, int M, mpf_class *A, int LDA);
void printsymmmat (int N, mpf_class *A, int LDA);
void printvec (int N, mpf_class *A);
void setrandsymmmat (int N, mpf_class *p);
void setzerovec (int N, double *p);
void printnum(double a);
void printnum(mpf_class a);
void printnum(complex<double> a);
void printnum(mpc_class a);

//implementation of debug support functions
mpc_class mpc_randomnumber(void)
{
  mpc_class ctmp;
  mpf_class mtmp1;
  mpf_class mtmp2;
  mtmp1 = uniformrandom->get_f();
  mtmp1 = 2.0 * mtmp1 - 1.0;

  mtmp2 = uniformrandom->get_f();
  mtmp2 = 2.0 * mtmp2 - 1.0;

  ctmp.re = mtmp1;
  ctmp.im = mtmp2;
  return ctmp;
}

mpf_class mpf_randomnumber(void)
{
  mpf_class mtmp;
  mtmp = uniformrandom->get_f();
//  mtmp = 2.0 * mtmp - 1.0;
  return mtmp;
}

void setzerovec (int N, mpf_class *p)
{
  if(N<=0) return;
  for (int i=0; i<N; i++){
    p[i] = 0.0;
  }
}

void setzerovec (int N, double *p)
{
  if(N<=0) return;
  for (int i=0; i<N; i++){
    p[i] = 0.0;
  }
}

void setrandvec (int N, mpf_class *p)
{
  if(N<=0) return;
  for (int i=0; i<N; i++){
    p[i] = mpf_randomnumber();
  }
}

void setrandvec (int N, mpc_class *p)
{
  if(N<=0) return;
  for (int i=0; i<N; i++){
    p[i] = mpc_randomnumber();
  }
}

//set random values in a matrix
void setrandmat (int N, int M, mpf_class *p)
{
  mpf_class mtmp;
  for (int j=0; j<M; j++){
    for (int i=0; i<N; i++){
      mtmp = mpf_randomnumber();
        p[i+j*N] = mtmp;
    }
  }
}

//print matrix as matlab compat format.
void printmat (int N, int M, mpf_class *A, int LDA)
{
  mpf_class mtmp;
  printf("[ ");
  for (int i=0; i<N; i++){
    printf("[ ");
    for (int j=0; j<M; j++){
      mtmp=A[i+j*LDA];
      gmp_printf( GMP_P_FORMAT, mtmp.get_mpf_t());
      if (j<M-1) printf(", ");
    }
    if (i<N-1) printf("]; "); else printf ("] ");
  }
  printf("]");
}

void printmat (int N, int M, mpc_class *A, int LDA)
{
  mpc_class mtmp;
  printf("[ ");
  for (int i=0; i<N; i++){
    printf("[ ");
    for (int j=0; j<M; j++){
      mtmp=A[i+j*LDA];
      gmp_printf( GMP_P_FORMAT "+" GMP_P_FORMAT "i", mtmp.re.get_mpf_t() , mtmp.im.get_mpf_t() );
      if (j<M-1) printf(", ");
    }
    if (i<N-1) printf("]; "); else printf ("] ");
  }
  printf("]");
}

void printmat (int N, int M, complex<double> *A, int LDA)
{
  complex<double> ctmp;
  printf("[ ");
  for (int i=0; i<N; i++){
    printf("[ ");
    for (int j=0; j<M; j++){
      ctmp=A[i+j*LDA];
      printf( P_FORMAT "+" P_FORMAT "i", ctmp.real(), ctmp.imag());
      if (j<M-1) printf(", ");
    }
    if (i<N-1) printf("]; "); else printf ("] ");
  }
  printf("]");
}

//print symmetry matrix as matlab compat format.
void printsymmmat (int N, mpf_class *A, int LDA)
{
  mpf_class mtmp;
  printf("[ ");
  for (int i=0; i<N; i++){
    printf("[ ");
    for (int j=0; j<N; j++){
      mtmp=A[i+j*LDA];
      gmp_printf(GMP_P_FORMAT, mtmp.get_mpf_t());
      if (j<N-1) printf(", ");
    }
    if (i<N-1) printf("]; "); else printf ("] ");
  }
  printf("]");
}

void printvec (int N, mpf_class *A)
{
  printf(" [ ");
  if(N!=0){
    for (int i=0; i<N-1; i++){
      gmp_printf( GMP_P_FORMAT "; ", A[i].get_mpf_t());
    }
    gmp_printf( GMP_P_FORMAT " ] ", A[N-1].get_mpf_t());
  }
  else gmp_printf( " ] ");
}

void printvec (int N, double *A)
{
  printf(" [ ");
  if(N!=0){
    for (int i=0; i<N-1; i++){
      printf( P_FORMAT "; ", A[i]);
    }
    printf( P_FORMAT " ] ", A[N-1]);
  } else printf( " ] ");
}

void printvec (int N, complex<double> *A)
{
  printf(" [ ");
  if(N!=0){
    for (int i=0; i<N-1; i++){
      printf( P_FORMAT "+" P_FORMAT "i; ", A[i].real(), A[i].imag());
    }
    printf( P_FORMAT "+" P_FORMAT "i ] ", A[N-1].real(), A[N-1].imag());
  } else printf( " ] ");
}

void printvec (int N, mpc_class *A)
{
  printf(" [ ");
  if(N!=0){
    for (int i=0; i<N-1; i++){
      gmp_printf( GMP_P_FORMAT "+" GMP_P_FORMAT "i; ", A[i].re.get_mpf_t() , A[i].im.get_mpf_t() );
    }
    gmp_printf( GMP_P_FORMAT "+" GMP_P_FORMAT "i ] ", A[N-1].re.get_mpf_t(), A[N-1].im.get_mpf_t());
  }
  else gmp_printf( " ] ");
}


void printnum(double a)
{
     printf( P_FORMAT " ", a);
}
void printnum(mpf_class a)
{
     gmp_printf( GMP_P_FORMAT " ", a.get_mpf_t());
}

void printnum(complex<double> a)
{
     printf( P_FORMAT "+" P_FORMAT "i ", a.real(), a.imag());
}
void printnum(mpc_class a)
{
     gmp_printf( GMP_P_FORMAT "+" GMP_P_FORMAT "i " , a.re.get_mpf_t(), a.im.get_mpf_t());
}

//set random symmetric matrix.
void setrandsymmmat (int N, mpf_class *p)
{
  mpf_class mtmp;
  for (int i=0; i<N; i++){
    for (int j=i; j<N; j++){
      mtmp = mpf_randomnumber();
      p[i+j*N] = mtmp;
      p[j+i*N] = mtmp;
    }
  }
}

void set_random_number(mpc_class *alpha, complex<double> *alpha_d)
{
  *alpha=mpc_randomnumber();
  alpha_d->real()=alpha->re.get_d();
  alpha_d->imag()=alpha->im.get_d();
}

void set_random_number(mpf_class *alpha, double *alpha_d)
{
  *alpha=mpf_randomnumber();
  *alpha_d=(*alpha).get_d();
}

void set_random_vector(mpc_class *vec, complex<double> *vec_d, int len)
{
  if(len<=0) return;
  for(int i=0;i<len;i++){
    vec[i] = mpc_randomnumber();
    vec_d[i].real()=vec[i].re.get_d();
    vec_d[i].imag()=vec[i].im.get_d();
  }
}

void set_random_vector(mpf_class *vec, double *vec_d, int len)
{
  if(len<=0) return;
  for(int i=0;i<len;i++){
    vec[i] = mpf_randomnumber();
    vec_d[i]=vec[i].get_d();
  }
}

double vec_diff(mpc_class *vec, complex<double> *vec_d, int len, int inc)
{
  double dtemp;
  double diff=0.0;

  for(int i=0;i<len*inc;i=i+inc){
    dtemp = vec[i].re.get_d()-vec_d[i].real();
    diff=diff+fabs(dtemp);
    dtemp = vec[i].im.get_d()-vec_d[i].imag();
    diff=diff+fabs(dtemp);
  }
  return diff;
}

double vec_diff(mpf_class *vec, double *vec_d, int len, int inc)
{
  double dtemp;
  double diff=0.0;

  for(int i=0;i<len*inc;i=i+inc){
    dtemp = vec[i].get_d()-vec_d[i];
    diff=diff+fabs(dtemp);
  }
  return diff;
}


double num_diff(mpc_class num1, complex<double> num2)
{
  double dtemp;
  double diff;
  dtemp = num1.re.get_d()-num2.real();
  dtemp = num1.im.get_d()-num2.imag();
  return dtemp;
}

double num_diff(mpf_class num1, double num2)
{
  double dtemp;
  double diff;
  dtemp = num1.get_d()-num2;
  return dtemp;
}
